% LSQFIT_DL  Simple implementation of trust-region dogleg method for
% non-linear regression.
%
% Usage:
%   [b, res, J] = lsqfit_dl(X, Y, func, b0)
% Inputs:
%   X -- independent variable(s) NxM matrix for function func
%   Y -- dependent variable Nx1 vector for function func
%   func --- anonymous function: [F, J] = func(b, X)
%            F and J are Nx1 func.values and NxM Jacobian respectively
%            b is Mx1 function parameters vector
% Outputs:
%   b -- optimized coefficients
%   res -- Nx1 regression residuals vector
%   J -- NxM Jacobian of the function for b and X values
%
% See also LSQFIT_LM, LSQFIT_GN, LSQFIT_CI
function [b, res, J] = lsqfit_dl(X, Y, func, b0)
% Check input parameters sizes
b0 = b0(:); Y = Y(:); % Transform to b0 and Y to column vectors
if size(X, 1) == 1
    X = X';
end
if size(X, 1) ~= numel(Y)
    error('X and Y sizes are not consistent');
end
% Some initial parameters
disp('Trust-region dogleg method based non-linear regression solver');
verbose = false; % Show D changes inside one step (function value)
delta = 1e-12; % Solver accuracy (b tolerance)
ftol = 1e-12; % Solver accuracy (function tolerance)
D = 10; Dmin = 1e-12; Dmax = 1e6; % Current, min and max trust region radii
m = numel(b0); % Number of parameters
db = 1; dF = 1; % Initial values to run initial iteration
[f, J] = func(b0, X); f = f - Y; % Initial function value (and Jacobian)
% Method iterations
inside_step = false;
while D > Dmin || (sqrt((db'*db)/m) > delta && sqrt(dF / m) > ftol)
    % Show the initial approximation
    if verbose || ~inside_step
        fprintf('b=[%s] F=%10.4g ', sprintf('%10.5g ', b0), f'*f);
        inside_step = true;
    end
    % Cauchy step (steepest descent) estimation
    g = J'*f;
    db_sd = -(g'*g) ./ ((J*g)'*J*g) .* g;
    sd_norm = sqrt(db_sd'*db_sd);    
    % Selection of total dogleg step
    if sd_norm > D
        % Steepest gradient step estimation
        db_dl = db_sd / sd_norm * D;
        step_type = 'SD';
    else
        % Gauss-Newton step estimation
        h = J'*J; % Note: Hessian is 2*J'*J
        if rcond(h) < eps % Hessian is singular: do steepest gradient step
            db_dl = db_sd; step_type = 'SDIN';
        else % Hessian is not singular: do Gauss-Newton step
            db_gn = - h \ g;            
            if sqrt(db_gn'*db_gn) < D % GN step is inside the trust region
                db_dl = db_gn; step_type = 'GN';
            else % GN step is outside the trust region: truncate
                a = db_sd; b = db_gn - db_sd; % (quadratic equation)
                ab = a'*b; aa = a'*a; bb = b'*b;
                w = (-ab + sqrt(ab.^2 - bb*(aa - D.^2))) ./ bb;
                db_dl = db_sd + w * (db_gn - db_sd);
                step_type = 'GNTR';
                %draw_dl_step(b0, D, db_sd, db_gn, db_dl);
            end
        end
    end    
    % Estimation of rho value (the new approximation quality)
    b0_new = b0 + db_dl;
    [fnew, Jnew] = func(b0_new, X); fnew = fnew - Y;
    dF = f'*f - fnew'*fnew;
    dL = -(2*g'*db_dl + (J*db_dl)'*(J*db_dl));    
    rho = dF ./ dL;
    % Select new approximation
    if rho > 0
        b0 = b0_new; f = fnew; J = Jnew; % Accept new b value
        inside_step = false;
    end
    % Show rho and step type
    if verbose || ~inside_step
        fprintf('D=%10.4g r=%10.3g (%s)\n', D, rho, step_type);
    end
    % Select new trust region radius
    if rho <= 0.25;    D = D / 4; % Decrease t.r. radius
    elseif rho > 0.75; D = min(D * 2, Dmax); % Increase t.r. radius
    end
end
% Show and return the result
if ~verbose && inside_step
    fprintf('\n');
end
fprintf('b=[%s] F=%10.4g D=%10.4g r=%10.3g (%s)\n', ...
    sprintf('%10.5g ', b0), f'*f, D, rho, step_type);
b = b0; res = f;
end
