% LSQFIT_EX Example of Gauss-Newton and Levenberg-Marquardt methods
% implementation. Data set for fitting is taken from:
% Graybill F.A. and Iyer H.K. Regression analysis. Concepts and
% applications. 1994. Duxbury Pr. ISBN 0534198694
% http://www.stat.colostate.edu/regression_book/chapter9.pdf
%
% You can launch the file either in MATLAB or in GNU Octave
%
% See also LSQFIT_LM, LSQFIT_GN, LSQFIT_EX
function lsqfit_ex
% Input data set + some settings
data = [ ...
    1    2    3    4    5    6    7    8    9    10   11   12
    2.86 2.64 1.57 1.24 0.45 1.02 0.65 0.18 0.15 0.01 0.04 0.36
    0.0  0.0  1.0  1.0  2.0  2.0  3.0  3.0  4.0  4.0  5.0  5.0];
X = data(3,:)'; % Concentration, mg/l
Y = data(2,:)'; % Optical reading

% Set initial approximation
%b0 = [10 6 2.5]; % No GN method F rise
%b0 = [0 3 2]; % GN method F rise
b0 = [10 10 10]; %GN method doesn't converge
% Call solver(s)
a = menu('Select the method', ...
    'Gauss-Newton', 'Levenberg-Marquardt', ...
    'Trust-region dogleg', 'lsqcurvefit');
if a == 1
    [b, res, J] = lsqfit_gn(X, Y, @func, b0);
elseif a == 2
    [b, res, J] = lsqfit_lm(X, Y, @func, b0);
elseif a == 3
    [b, res, J] = lsqfit_dl(X, Y, @func, b0);
elseif a == 4
    % Requires either GNU Octave optim package
    % or MATLAB Optimization Toolbox
    opt = optimset('Display', 'iter', 'Jacobian', 'on');
    [b,~,res,~,~,~,J] = lsqcurvefit(@func, b0, X, Y, [], [], opt);
    J = full(J);
end

% Calculate and show confidence intervals
[db, b_lb, b_ub, sb] = lsqfit_ci(b, res, J);
disp('Regression results');
for i=1:numel(b)
    fprintf('b(%d)=%.5f sb=%.3f db=%.3f b=%8.5f-%8.5f\n', ...
        i, b(i), sb(i), db(i), b_lb(i), b_ub(i));
end
% Show graph
xv = 0:0.1:5;
figure;
hold on;
plot(X, Y, 'bo', 'LineWidth', 2);
plot(xv, func(b,xv), 'k-', 'LineWidth', 2);
hold off;
legend('Data', 'Regression');
xlabel('c, mg/l'); ylabel('\mu');
end

function [F, J] = func(b, x)
F = b(1) + b(2)*exp(-b(3)*x); % Function
if nargout == 2 % and its Jacobian
    df_db1 = ones(size(x));
    df_db2 = exp(-b(3)*x);
    df_db3 = -b(2)*exp(-b(3)*x).*x;
    J = [df_db1 df_db2 df_db3];
end
end