% LSQFIT_GN  Simple implementation of Gauss-Newton method for
% non-linear regression.
%
% Usage:
%   [b, res, J] = lsqfit_gn(X, Y, func, b0)
% Inputs:
%   X -- independent variable(s) NxM matrix for function func
%   Y -- dependent variable Nx1 vector for function func
%   func --- anonymous function: [F, J] = func(b, X)
%            F and J are Nx1 func.values and NxM Jacobian respectively
%            b is Mx1 function parameters vector
% Outputs:
%   b -- optimized coefficients
%   res -- Nx1 regression residuals vector
%   J -- NxM Jacobian of the function for b and X values
%
% See also LSQFIT_LM, LSQFIT_DL, LSQFIT_CI
function [b, res, J] = lsqfit_gn(X, Y, func, b0)
% Check input parameters sizes
Y = Y(:); b = b0(:); % Convert Y and b0 to column vectors
if size(X, 1) == 1
    X = X';
end
if size(X, 1) ~= numel(Y)
    error('X and Y sizes are not consistent');
end
% Some initial parameters
disp('Gauss-Newton method based non-linear regression solver');
delta = 1e-15; % Solver accuracy
db = 1; % Initial value (to run initial iteration)
% Iterations
while sqrt((db'*db)/numel(db)) > delta    
    % Calculate and show initial approximation    
    [F, J] = func(b, X); res = F(:) - Y;
    fprintf('b=[%s] F=%10.4g\n', sprintf('%10.4g ', b), res'*res);
    % Process iteration
    h = J'*J; c = rcond(h);
    if isnan(c) || c < eps
        error('Ill-conditioned equation system in the iteration');
    end
    db = h \ (-J'*res);
    if any(isnan(db) | isinf(db))
        error('Ill-conditioned equation system in the iteration');
    end
    b = b + db;
end
% Show the result
fprintf('b=[%s] F=%10.4g\n', sprintf('%10.4g ', b), res'*res);
end