% LSQFIT_LM  Simple implementation of Levenberg-Marquardt method for
% non-linear regression.
%
% Usage:
%   [b, res, J] = lsqfit_lm(X, Y, func, b0)
% Inputs:
%   X -- independent variable(s) NxM matrix for function func
%   Y -- dependent variable Nx1 vector for function func
%   func --- anonymous function: [f, J] = func(b, X)
%            f and J are Nx1 func.values and NxM Jacobian respectively
%            b is Mx1 function parameters vector
% Outputs:
%   b -- optimized coefficients
%   res -- Nx1 regression residuals vector
%   J -- NxM Jacobian of the function for b and X values
%
% See also LSQFIT_GN, LSQFIT_DL, LSQFIT_CI
function [b, res, J] = lsqfit_lm(X, Y, func, b0)
% Check input parameters sizes
Y = Y(:); b0 = b0(:); % Convert b0 and Y to column vectors
if size(X, 1) == 1
    X = X';
end
if size(X, 1) ~= numel(Y)
    error('X and Y sizes are not consistent');
end
% Some initial parameters
delta = 1e-15; % Solver accuracy (b and function tolerances)
ftol = 1e-15; % Solver accuracy (function tolerance)
l = 1e-8; % Initial value of the lambda regularization parameter
nu = 10; % Scaling factor for lambda parameter search
% Calculate Jacobian and initial function value
[f, J] = func(b0, X); f = f - Y;
% Levenberg-Marquardt method cycle
m = numel(b0); % Number of parameters
db = 1; dF = 1; % Initial values (to run initial iteration)
disp('Levenberg-Marquardt method based non-linear regression solver');
while sqrt((db'*db)/m) > delta && sqrt(dF / m) > ftol
    % Show initial approximation
    fprintf('b=[%s] F=%10.4g l=%.3e\n', sprintf('%10.4g ', b0), f'*f, l);
    % LM algorithm step finding
    l = l / nu^2; % Initial step divided by nu
    fnew = 1.1 * f; % Initial value to run the cycle
    while fnew'*fnew > f'*f
        l = l * nu; % Increase step        
        A = J'*J + l*eye(m);
        if rcond(A) > eps
            db = -A \ (J'*f);
            if any(isnan(db) | isinf(db))
                error('Ill-conditioned equation system in the iteration');
            end        
            b0_new = b0 + db;
            [fnew, Jnew] = func(b0_new, X); fnew = fnew - Y;
        end
    end
    % Accept new approximation
    dF = f'*f - fnew'*fnew; b0 = b0_new; f = fnew; J = Jnew;
end
% Show and return result
fprintf('b=[%s] F=%10.4g l=%.3e\n', sprintf('%10.4g ', b0), f'*f, l);
b = b0; res = f;
end
